"""I/O functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_util.ipynb.

# %% auto 0
__all__ = ['read_gff', 'gff_attribute_selector', 'decorate', 'filter', 'syntenic_block_borders', 'read_aln', 'estimate_plot_size',
           'insert_gap', 'flip', 'insert_break']

# %% ../nbs/01_util.ipynb 4
import os
from collections.abc import Sequence
from pathlib import Path
from typing import Literal

import numpy as np
import pandas as pd

# %% ../nbs/01_util.ipynb 6
def read_gff(
    loc: str | Path,  # input filepath
    gff_columns: list = [
        "seqid",
        "source",
        "type",
        "start",
        "end",
        "score",
        "strand",
        "phase",
        "attributes",
    ],  # column names for the GFF3 file
    skiprows: int = 1,  # how many rows to skip in the beginning
    header: (
        int | Sequence[int] | None | Literal["infer"]
    ) = None,  # whether to expect a header or not
    sep: str = "\t",  # separator for the table
    **kwargs,  # various pd.read_csv() arguments
) -> pd.DataFrame:  # the GFF3 file in DataFrame form
    "A function to read a GFF3 file. Expects 9 tab-separated fields, and will try to name the columns according to https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md"
    gff = pd.read_csv(loc, sep=sep, header=header, skiprows=skiprows, **kwargs)
    gff.columns = gff_columns
    return gff

# %% ../nbs/01_util.ipynb 11
def gff_attribute_selector(
    line,  # the attributes field of a GFF3 line
    sep: str = ";",  # the field separator. Should be a semicolon for a GFF3 file.
    select="ID",  # the field ID. Should be one that is included in the GFF3 file. Refer to https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md for possibilities, or choose a manually defined tag that you know is present in the file.
) -> str:  # the value of field `select`
    "A function that extracts the value of a specified field from the attributes of a GFF3 line. Should be used on the `attributes` field of the corresponding pandas DataFrame."
    for field in line.split(sep):
        if field.startswith(select):
            value = field.split("=")[1]
            return value
    return None

# %% ../nbs/01_util.ipynb 16
def decorate(
    gff: pd.DataFrame,  # a GFF file in Pandas dataframe form
    attributes: dict = {
        "gene_id": "ID",
        "gene_name": "gene",
    },  # a dictionary of tags to extract from the attributes column. For each key/value pair, the value under the `key` tag will be saved in a new column named `value`.
):
    "A function that"
    for attr_add, attr_filter in attributes.items():
        gff[attr_add] = gff["attributes"].apply(
            lambda x: gff_attribute_selector(x, select=attr_filter)
        )


# | export


def filter(
    gff,
    filter_by_type=True,
    filter_type="gene",
    filter_by_field=True,
    field="gene_id",
    field_values=None,
):
    if filter_type and filter_type not in gff[filter_type]:
        raise ValueError(
            f'The value {filter_type} is not found in the "type" column of the input table.'
        )

    if filter_by_field:
        if field not in gff.columns:
            raise ValueError(
                f'The value {filter_type} is not found in the "type" column of the input table.'
            )

    keep = np.array([True] * gff.shape[0])
    if genes is not None:
        is_genes = gff["gene_id"].isin(genes)
        keep = keep & is_genes

    if is_type:
        is_gene = gff["type"] == "gene"

    slim = gff[is_gene & is_hox]
    return slim

# %% ../nbs/01_util.ipynb 17
def syntenic_block_borders(
    gff: pd.DataFrame,  # a GFF in Pandas dataframe form. Only includes the genes of the syntenic block in question.
    flank_length: int = None,  # the amount of space to be granted on both sides of the syntenic region, in basepairs. If unspecified, it will be set to 5% of the syntenic block length.
    start: str = "start",  # the GFF column with the start position of the gene ("start").
    end: str = "end",  # the GFF column with the end position of the gene ("end").
) -> (int, int):
    "A function to calculate the boundaries of a syntenic block. It automatically pads the boundaries by an additional 5% of total length on both ends."
    possible = [gff[start].min(), gff[start].max(), gff[end].min(), gff[end].max()]
    block_start = min(possible)
    block_end = max(possible)
    if flank_length is None:
        flank_length = (block_end - block_start) // 100 * 5
    if not flank_length.is_integer():
        raise ValueError(
            f'The parameter `flank_length` has to be an integer. You supplied the value "{flank_length}" which is {type(flank_length)}.'
        )
    return max(block_start - flank_length, 0), block_end + flank_length

# %% ../nbs/01_util.ipynb 20
def read_aln(
    m8: str,  # the path to the MMseqs2 alignment table file
    id_sep: (
        str | None
    ) = None,  # (optional) a character that separates the species ID from the gene ID
    **kwargs,  # various arguments to be passed to the pd.read_csv() function
) -> pd.DataFrame:  # the tabulated form of the alignment results.
    "Reads"
    hits = pd.read_csv(m8, sep="\t", header=None)
    m8_columns = [
        "query",
        "target",
        "seq_id",
        "ali_len",
        "no_mism",
        "no_go",
        "q_start",
        "q_end",
        "t_start",
        "t_end",
        "eval",
        "bit",
    ]
    hits.columns = m8_columns
    # trim the query to just the ID
    if id_sep is not None:
        hits["query"] = hits["query"].str.split(id_sep).str[1].str.join(id_sep)
    return hits

# %% ../nbs/01_util.ipynb 21
def estimate_plot_size(
    gff,
    width_factor: int = 3,
    height: int = 2,
):
    x = len(gff) * width_factor
    return (x, height)

# %% ../nbs/01_util.ipynb 23
def insert_gap(
    gff: pd.DataFrame,
    locus1=None,
    locus2=None,
    identifier="gene_id",
    purge_columns=None,
    no_gaps=1,
) -> pd.DataFrame:
    "This function inserts a number of dummy entries between two loci (lines) in the GFF DataFrame."
    loc1_index = -1
    loc2_index = len(gff)
    if locus1 is not None:
        loc1_index = gff[gff[identifier] == locus1].index[0]
    if locus2 is not None:
        loc2_index = gff[gff[identifier] == locus2].index[0]
    if loc2_index - loc1_index != 1:
        raise ValueError(
            f"The two loci are not consecutive; their indices are {locus1}: {loc1_index} and {locus2}: {loc2_index}, respectively. Please refer to your input dataframe."
        )

    offset_step = 1 / (no_gaps + 1)

    copy = loc1_index
    if locus1 is None:
        copy = loc2_index

    for i in range(no_gaps):
        offset = (i + 1) * offset_step
        tmp = gff.loc[copy].copy()
        if locus1 is not None:
            tmp["start"] = tmp["end"] + 1 + (i * 2)
            tmp["end"] = tmp["start"] + 1
        else:
            tmp["start"] = tmp["start"] - 1 - (i * 2)
            tmp["end"] = tmp["start"] + 1

        tmp[identifier] = f"gap_{locus1}-{i}"
        for column in purge_columns:
            tmp[column] = ""
        gff.loc[loc1_index + offset] = tmp

    gff = gff.sort_index().reset_index(drop=True)
    return gff

# %% ../nbs/01_util.ipynb 29
def _flip_strand(s):
    if s == "+":
        return "-"
    elif s == "-":
        return "+"
    else:
        return s


# | export


def flip(gff):
    gff["start"] = (-1) * gff["start"]
    gff["end"] = (-1) * gff["end"]
    gff["strand"] = gff["strand"].apply(_flip_strand)
    gff.index = gff.index[::-1]
    return gff.sort_index()

# %% ../nbs/01_util.ipynb 31
def insert_break(
    gff: pd.DataFrame,
    locus1=None,
    locus2=None,
    identifier="gene_id",
) -> pd.DataFrame:
    "This function inserts a molecule break between two loci (lines) in the GFF DataFrame."
    loc1_index = gff[gff[identifier] == locus1].index[0]
    loc2_index = gff[gff[identifier] == locus2].index[0]
    if loc2_index - loc1_index != 1:
        raise ValueError(
            f"The two loci are not consecutive; their indices are {locus1}: {loc1_index} and {locus2}: {loc2_index}, respectively. Please refer to your input dataframe."
        )

    tmp = gff.loc[loc1_index].copy()
    for c in tmp.index:
        tmp[c] = ""
    tmp[identifier] = f"break"
    gff.loc[loc1_index + 0.5] = tmp

    gff = gff.sort_index().reset_index(drop=True)
    return gff



"""The plotting functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['plot_synteny', 'plot_synteny_schematic']

# %% ../nbs/00_core.ipynb 4
import os
from pathlib import Path
from typing import Tuple, Union

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import pandas as pd
from matplotlib import pyplot as plt

from . import util

# %% ../nbs/00_core.ipynb 8
def plot_synteny(
    gff,  # a GFF in Pandas dataframe form. Only includes the genes of the syntenic block in question.
    block_start: int = None,  # The start coordinate of the syntenic locus to plot. If None, will be set to the start of the first gene minus 5% of the block length.
    block_end: int = None,  # The end coordinate of the syntenic locus to plot. If None, will be set to the end of the last gene plus 5% of the block length.
    figsize: (float, float) = None,  # the figure size.
    locus_start: str = "start",  # the GFF column that describes the start coordinate of the entities (e.g. gene, mRNA) to be plotted on the chromosome. In a well-behaved GFF, this should be "start".
    locus_end: str = "end",  # the GFF column that describes the end coordinate of the entities (e.g. gene, mRNA) to be plotted on the chromosome. In a well-behaved GFF, this should be "end".
    molecule: str = "seqid",  # the GFF column that holds the molecule name (chromosome/scaffold/contig ID). In a well-behaved GFF, this should be "seqid".
    locus_name: str = "gene_name",  # the GFF column that holds the gene/mRNA symbol. This is usually a tag in the "attributes" column of a well-behaved GFF and should have been extracted prior to this step. It will be plotted according to the `gene_name_offset` parameter.
    fig_title: str = None,  # Figure title.
    save: str | Path = "synteny.svg",  # path to save an SVG image.
    palette: (
        dict | None
    ) = mcolors.CSS4_COLORS,  # color palette in dictionary form. If None, the color column is expected to be a string that can be interpreted as color.
    chromosome_color: str = "lightgray",  # the color of the line representing the molecule the genes are plotted on.
    backup_gene_color: str = "darkgray",  # If the GFF has no column titled "color", the gene arrows will be filled using this color.
    gene_name_offset: float = 0.02,  # Default offset for plotting gene names if the GFF has no column titled "offset".
    chromosome_width: float = 1,  # Linewidth parameter for the line representing the chromosome.
    gene_width: float = 5,  # Width of the arrow patches representing genes. Please refer to the documentation of `matplotlib.pyplot.arrow` for more details.
) -> None:
    "Function to plot a syntenic cluster of genes."

    if len(gff[molecule].unique()) > 1:
        raise ValueError(
            f"There are multiple molecules represented in the input GFF: {gff[molecule].unique()}. Please only input a GFF with genes on a single molecule."
        )

    if block_start is None or block_end is None:
        block_start, block_end = util.syntenic_block_borders(gff)

    if figsize is None:
        figsize = util.estimate_plot_size(gff)

    fig, ax = plt.subplots(figsize=figsize)

    # plot contig
    ax.plot(
        [block_start, block_end],
        [0, 0],
        color=chromosome_color,
        linewidth=chromosome_width,
    )
    for name, gene in gff.iterrows():
        start = gene[locus_start]
        end = gene[locus_end]
        if "color" in gene.index:
            if palette is not None:
                color = mcolors.CSS4_COLORS[gene["color"]]
            else:
                color = gene["color"]
        else:
            color = backup_gene_color

        if "offset" in gene.index:
            offset = gene["offset"]
        else:
            offset = gene_name_offset

        if gene["strand"] == "-":
            arrow_from = end
            arrow_to = start - end - 1000
        else:
            arrow_from = start
            arrow_to = end - start + 1000

        ax.arrow(
            arrow_from,
            0,
            arrow_to,
            0,
            head_width=0.05,
            head_length=1000,
            fc=color,
            ec=color,
            linewidth=gene_width,
        )
        ax.text(
            gene[locus_start] - 10000,
            offset * 20,
            gene[locus_name],
            color=color,
            fontsize=10,
        )

    ax.set_yticks([])
    ax.set_ylim(-1.5, 1)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    if fig_title is not None:
        ax.set_title(fig_title)
    plt.gca().set_position([0, 0, 1, 1])
    if save is not None:
        if not (isinstance(save, str) | isinstance(save, Path)):
            raise TypeError(
                f"The save location should be a valid Path or string; instead, it is {type(save)}."
            )
        plt.savefig(save, bbox_inches="tight", pad_inches=0)

# %% ../nbs/00_core.ipynb 17
def plot_synteny_schematic(
    gff: pd.DataFrame,  # a GFF in Pandas dataframe form. Only includes the genes of the syntenic block in question. Assumed to be sorted in plotting order.
    block_gene: float = 400,  # length of a gene arrow, in plot coordinate space.
    block_dist: float = 200,  # length of gap between successive genes, in plot coordinate space.
    locus_start: str = "start",  # the GFF column that describes the start coordinate of the entities (e.g. gene, mRNA) to be plotted on the chromosome. In a well-behaved GFF, this should be "start".
    locus_end: str = "end",  # the GFF column that describes the end coordinate of the entities (e.g. gene, mRNA) to be plotted on the chromosome. In a well-behaved GFF, this should be "end".
    molecule: str = "seqid",  # the GFF column that holds the molecule name (chromosome/scaffold/contig ID). In a well-behaved GFF, this should be "seqid".
    locus_name: str = "gene_name",  # the GFF column that holds the gene/mRNA symbol. This is usually a tag in the "attributes" column of a well-behaved GFF and should have been extracted prior to this step. It will be plotted according to the `gene_name_offset` parameter.
    interrupted: bool = False,  # is the syntenic cluster on multiple pseudomolecules?
    palette: (
        dict | None
    ) = mcolors.CSS4_COLORS,  # color palette in dictionary form. If None, the color column is expected to be a string that can be interpreted as color.
    fontsize: int = 30,  # this fontsize works well for default sizes, but is probably too big for smaller plots
    chromosome_color: str = "black",  # the color of the line representing the molecule the genes are plotted on.
    backup_gene_color: str = "darkgray",  # If the GFF has no column titled "color", the gene arrows will be filled using this color.
    gene_name_offset: float = 0.02,  # Default offset for plotting gene names if the GFF has no column titled "offset".
    chromosome_width: float = 3,  # Linewidth parameter for the line representing the chromosome; also, patch linewidth of the arrows representing genes. Please refer to the documentation of `matplotlib.patches.FancyArrowPath` for more details.
    head_length: float = 200,  # Head length for the arrows representing genes. Please refer to the documentation of `matplotlib.patches.FancyArrowPath` for more details.
    head_width: float = 1.7,  # Head width for the arrows representing genes. Please refer to the documentation of `matplotlib.patches.FancyArrowPath` for more details.
    gene_width: float = 1,  # Width of the arrow patches representing genes. Please refer to the documentation of `matplotlib.patches.FancyArrowPath` for more details.
    ax: plt.Axes = None,  # a pre-existing matplotlib axes object, if you wish to include this plot in a composite.
    figsize: (
        float,
        float,
    ) = None,  # the figure size. Will be ignored if ax is not None.
    save: str | Path = "schematic.svg",  # path to save an SVG image.
    return_fig: bool = False,  # whether to return an axes object. Useful for further manipulation of the plot.
) -> plt.Figure | plt.Axes | None:
    # 'Function to plot a syntenic cluster of genes. Each gene is represented by an arrow; left-facing arrows indicate a minus orientation, and right-facing arrows a plus orientation. The genes can be colored using named colors from the `matplotlib.colors.CSS4_COLORS` collection.'

    if len(gff[molecule].unique()) > 1:
        if not interrupted:
            raise ValueError(
                f"There are multiple molecules represented in the input GFF: {gff[molecule].unique()}. Please only input a GFF with genes on a single molecule."
            )

    if figsize is None:
        figsize = util.estimate_plot_size(gff)

    fig, ax = plt.subplots(figsize=figsize)
    # the total length is:
    # - the number of gene blocks
    # - the number of gaps between genes (number of gene blocks - 1)
    # - the start and end gap
    gff_genes = gff[gff[molecule] != ""]
    total_length = (
        (len(gff_genes) * block_gene)
        + (2 * block_dist)
        + ((len(gff_genes) - 1) * block_dist)
    )
    # plot contig
    ax.plot(
        [0, total_length], [0, 0], color=chromosome_color, linewidth=chromosome_width
    )
    start = 0
    end = 0
    genes_seen = 0
    for i, gene in gff.iterrows():
        if gene[molecule] != "":
            genes_seen += 1
        if gene[locus_start] == "":
            # as many gene blocks as we walked, plus as many gaps minus one
            break_start = (genes_seen) * block_gene + (genes_seen + 0.2) * block_dist
            break_end = break_start + 0.6 * block_dist
            ax.plot(
                [break_start, break_end],
                [0, 0],
                color="white",
                linewidth=chromosome_width + 2,
            )
            offset = gene_name_offset
            continue

        start = end + block_dist
        end = start + block_gene
        if "color" in gene.index:
            if gene["color"] == "":
                continue
            if palette is not None:
                color = mcolors.CSS4_COLORS[gene["color"]]
            else:
                color = gene["color"]
        else:
            color = backup_gene_color

        if "offset" in gene.index:
            offset = gene["offset"]
        else:
            offset = gene_name_offset

        # Draws an arrow from (x, y) to (x + dx, y + dy). The width of the arrow is scaled by width
        if gene["strand"] == "-":
            arrow_from = end
            arrow_to = start - end + head_length
        else:
            arrow_from = start
            arrow_to = end - start - head_length

        ec = "black"
        if "edge_color" in gene.index:
            ec = gene["edge_color"]

        arrow = mpatches.FancyArrow(
            arrow_from,
            0,
            arrow_to,
            0,
            width=gene_width,
            head_length=head_length,
            head_width=head_width,
            ec=ec,
            fc=color,
            lw=chromosome_width,
            zorder=2,
            head_starts_at_zero=False,
        )

        if gene["strand"] == ".":
            arrow_from = end
            arrow_to = start - end
            arrow = mpatches.FancyArrow(
                arrow_from,
                0,
                arrow_to,
                0,
                width=gene_width,
                head_length=0,
                head_width=0,
                ec=ec,
                fc=color,
                lw=chromosome_width,
                zorder=2,
                head_starts_at_zero=True,
            )

        ax.add_patch(arrow)
        # ax.arrow(end, 0, start-end-100, 0, head_width=0.05, head_length=100, fc=color, ec=color, linewidth=gene_width)
        ax.text(
            end - block_gene,
            offset * 60,
            gene[locus_name],
            color="black",
            fontsize=fontsize,
        )

    ax.set_yticks([])
    ax.set_xticks([])
    ax.set_ylim(-1.5, 1)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)

    plt.gca().set_position([0, 0, 1, 1])
    if save is not None:
        if not (isinstance(save, str) | isinstance(save, Path)):
            raise TypeError(
                f"The save location should be a valid Path or string; instead, it is {type(save)}."
            )
        plt.savefig(save, bbox_inches="tight", pad_inches=0)

    if return_fig:
        return ax



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/geneorder',
                'doc_host': 'https://galicae.github.io',
                'git_url': 'https://github.com/galicae/geneorder',
                'lib_path': 'geneorder'},
  'syms': { 'geneorder.core': { 'geneorder.core.plot_synteny': ('core.html#plot_synteny', 'geneorder/core.py'),
                                'geneorder.core.plot_synteny_schematic': ('core.html#plot_synteny_schematic', 'geneorder/core.py')},
            'geneorder.util': { 'geneorder.util._flip_strand': ('util.html#_flip_strand', 'geneorder/util.py'),
                                'geneorder.util.decorate': ('util.html#decorate', 'geneorder/util.py'),
                                'geneorder.util.estimate_plot_size': ('util.html#estimate_plot_size', 'geneorder/util.py'),
                                'geneorder.util.filter': ('util.html#filter', 'geneorder/util.py'),
                                'geneorder.util.flip': ('util.html#flip', 'geneorder/util.py'),
                                'geneorder.util.gff_attribute_selector': ('util.html#gff_attribute_selector', 'geneorder/util.py'),
                                'geneorder.util.insert_break': ('util.html#insert_break', 'geneorder/util.py'),
                                'geneorder.util.insert_gap': ('util.html#insert_gap', 'geneorder/util.py'),
                                'geneorder.util.read_aln': ('util.html#read_aln', 'geneorder/util.py'),
                                'geneorder.util.read_gff': ('util.html#read_gff', 'geneorder/util.py'),
                                'geneorder.util.syntenic_block_borders': ('util.html#syntenic_block_borders', 'geneorder/util.py')}}}
